﻿using Microsoft.VisualBasic.CompilerServices;
using System;
using System.Collections.Generic;
using System.Text;

namespace Raytracer
{
    public abstract class Shader
    {
        //return NonLightColor
        public abstract Vector3 Shade(Random rnd,Vector3 incident,Vector3 normal,out Vector3 exit,out Vector3 illum,Surface param = null);
    }

    public class DiffShader : Shader
    {
        Vector3 diffColor;
        public DiffShader(Vector3 color)
        {
            diffColor = color;
        }
        public override Vector3 Shade(Random rnd, Vector3 incident, Vector3 normal, out Vector3 exit, out Vector3 illum, Surface param = null)
        {
            Vector3 refx = normal.vertical.normalized;
            Vector3 refy = refx.Cross(normal).normalized;
            double theta = rnd.NextDouble() * Math.PI * 2;
            double fi = rnd.NextDouble();
            double fis = Math.Sqrt(fi);
            double x, y;
            x = Math.Sin(theta) * fis;
            y = Math.Cos(theta) * fis;
            double len = Math.Sqrt(1 - fi);
            exit = refx * x + refy * y + normal * len;
            illum = new Vector3(0);//normal / 2 + new Vector3(0.5);
            return diffColor;
            //return new Vector3(0);
        }
    }

    public class LightShader : Shader 
    {
        Vector3 col;
        public LightShader(Vector3 color)
        {
            col = color;
        }
        public override Vector3 Shade(Random rnd, Vector3 incident, Vector3 normal, out Vector3 exit, out Vector3 illum, Surface param = null)
        {
            illum = col;
            exit = new Vector3(0);
            return new Vector3(0);
        }
    }

    public class MirriorShader : Shader 
    {
        public override Vector3 Shade(Random rnd, Vector3 incident, Vector3 normal, out Vector3 exit, out Vector3 illum, Surface param = null)
        {

            illum = new Vector3(0);// normal / 2 + new Vector3(0.5);
            exit = incident - 2 * (incident.Dot(normal)) * normal;
            return new Vector3(1);
        }
    }
    public class DebugView : Shader
    {
        public override Vector3 Shade(Random rnd, Vector3 incident, Vector3 normal, out Vector3 exit, out Vector3 illum, Surface param = null)
        {

            illum =  incident / 2 + new Vector3(0.5);
            exit = incident - 2 * (incident.Dot(normal)) * normal;
            return new Vector3(0);
        }
    }
    public class DebugNormal : Shader
    {
        public override Vector3 Shade(Random rnd, Vector3 incident, Vector3 normal, out Vector3 exit, out Vector3 illum, Surface param = null)
        {

            illum = normal / 2 + new Vector3(0.5);
            exit = incident - 2 * (incident.Dot(normal)) * normal;
            return new Vector3(0);
        }
    }
    public class DebugExitDirection : Shader
    {
        public override Vector3 Shade(Random rnd, Vector3 incident, Vector3 normal, out Vector3 exit, out Vector3 illum, Surface param = null)
        {

            exit = incident - 2 * (incident.Dot(normal)) * normal;
            illum = exit / 2 + new Vector3(0.5);
            return new Vector3(0);
        }
    }
    public class DiffShaderA : Shader
    {
        Vector3 diffColor;
        double roughness;
        public DiffShaderA(Vector3 color,double roughness)
        {
            diffColor = color;
            this.roughness = roughness*roughness;
        }
        public override Vector3 Shade(Random rnd, Vector3 incident, Vector3 normal, out Vector3 exit, out Vector3 illum, Surface param = null)
        {
            incident = -incident;
            Vector3 pereExit = 2 * (incident.Dot(normal) * normal) - incident;
            Vector3 norx = pereExit.vertical.normalized;
            Vector3 nory = norx.Cross(pereExit).normalized;
            //do {
                double theta = rnd.NextDouble() * Math.PI * 2;
                double fi = rnd.NextDouble() * roughness;
                double fis = Math.Sqrt(fi);
                double x, y;
                x = Math.Sin(theta) * fis;
                y = Math.Cos(theta) * fis;
                double len = Math.Sqrt(1 - fi);
                exit = norx * x + nory * y + pereExit * len;
           // }
           // while (exit.Dot(normal) < 0);
            illum = new Vector3(0);
            return diffColor;
        }
    }

}
